import json

from django import forms
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wagtail.admin.staticfiles import versioned_static
from wagtail.admin.widgets import AdminChooser
from wagtail.telepath import register
from wagtail.widget_adapters import WidgetAdapter

from wagtailmedia.models import get_media_model


class AdminMediaChooser(AdminChooser):
    media_type = None
    choose_one_text = _("Choose a media item")
    choose_another_text = _("Choose another media item")
    link_to_chosen_text = _("Edit this media item")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.media_model = get_media_model()

    def get_value_data(self, value):
        if value is None:
            return value
        if not isinstance(value, self.media_model):
            value = self.media_model.objects.get(pk=value)
        return {
            "id": value.pk,
            "title": value.title,
            "edit_link": reverse("wagtailmedia:edit", args=[value.pk]),
        }

    def render_html(self, name, value, attrs):
        value_data = value if value is not None else {}

        original_field_html = super().render_html(name, value_data.get("id"), attrs)

        if self.media_type:
            chooser_url = reverse("wagtailmedia:chooser_typed", args=(self.media_type,))
        else:
            chooser_url = reverse("wagtailmedia:chooser")

        return render_to_string(
            "wagtailmedia/widgets/media_chooser.html",
            {
                "widget": self,
                "original_field_html": original_field_html,
                "attrs": attrs,
                "icon": "media",
                "value": value_data != {},  # only used to identify blank values
                "title": value_data.get("title", ""),
                "edit_url": value_data.get("edit_link", ""),
                "chooser_url": chooser_url,
            },
        )

    def render_js_init(self, id_, name, value):
        return f"createMediaChooser({json.dumps(id_)});"

    @property
    def media(self):
        return forms.Media(
            js=[
                "wagtailmedia/js/tabs.js",
                "wagtailmedia/js/media-chooser-modal.js",
                "wagtailmedia/js/media-chooser.js",
            ]
        )


class AdminAudioChooser(AdminMediaChooser):
    media_type = "audio"
    choose_one_text = _("Choose audio")
    choose_another_text = _("Choose another audio item")
    link_to_chosen_text = _("Edit this audio item")


class AdminVideoChooser(AdminMediaChooser):
    media_type = "video"
    choose_one_text = _("Choose video")
    choose_another_text = _("Choose another video")
    link_to_chosen_text = _("Edit this video")


class MediaChooserAdapter(WidgetAdapter):
    js_constructor = "wagtailmedia.MediaChooser"

    def js_args(self, widget):
        return [
            widget.render_html("__NAME__", None, attrs={"id": "__ID__"}),
            widget.id_for_label("__ID__"),
        ]

    @cached_property
    def media(self):
        return forms.Media(
            js=[
                versioned_static("wagtailmedia/js/media-chooser-telepath.js"),
            ]
        )


register(MediaChooserAdapter(), AdminMediaChooser)
register(MediaChooserAdapter(), AdminAudioChooser)
register(MediaChooserAdapter(), AdminVideoChooser)



import mimetypes
import os.path

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.validators import FileExtensionValidator, MinValueValidator
from django.db import models
from django.dispatch import Signal
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from wagtail.models import CollectionMember, ReferenceIndex
from wagtail.search import index
from wagtail.search.queryset import SearchableQuerySetMixin

from taggit.managers import TaggableManager

from wagtailmedia.settings import wagtailmedia_settings


ALLOWED_EXTENSIONS_THUMBNAIL = ["gif", "jpg", "jpeg", "png", "webp"]


class MediaType(models.TextChoices):
    AUDIO = "audio", _("Audio file")
    VIDEO = "video", _("Video file")


class MediaQuerySet(SearchableQuerySetMixin, models.QuerySet):
    pass


class AbstractMedia(CollectionMember, index.Indexed, models.Model):
    title = models.CharField(max_length=255, verbose_name=_("title"))
    file = models.FileField(upload_to="media", verbose_name=_("file"))

    type = models.CharField(
        choices=MediaType.choices, max_length=255, blank=False, null=False
    )
    duration = models.FloatField(
        blank=True,
        default=0,
        validators=[MinValueValidator(0)],
        verbose_name=_("duration"),
        help_text=_("Duration in seconds"),
    )
    width = models.PositiveIntegerField(null=True, blank=True, verbose_name=_("width"))
    height = models.PositiveIntegerField(
        null=True, blank=True, verbose_name=_("height")
    )
    thumbnail = models.FileField(
        upload_to="media_thumbnails", blank=True, verbose_name=_("thumbnail")
    )

    created_at = models.DateTimeField(verbose_name=_("created at"), auto_now_add=True)
    uploaded_by_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("uploaded by user"),
        null=True,
        blank=True,
        editable=False,
        on_delete=models.SET_NULL,
    )

    tags = TaggableManager(help_text=None, blank=True, verbose_name=_("tags"))

    objects = MediaQuerySet.as_manager()

    search_fields = CollectionMember.search_fields + [
        index.SearchField("title", partial_match=True, boost=10),
        index.FilterField("title"),
        index.RelatedFields(
            "tags",
            [
                index.SearchField("name", partial_match=True, boost=10),
            ],
        ),
        index.FilterField("uploaded_by_user"),
    ]

    admin_form_fields = (
        "title",
        "file",
        "collection",
        "duration",
        "width",
        "height",
        "thumbnail",
        "tags",
    )

    def __str__(self):
        return self.title

    @property
    def filename(self):
        return os.path.basename(self.file.name)

    @property
    def thumbnail_filename(self) -> str:
        return os.path.basename(self.thumbnail.name) if self.thumbnail else ""

    @property
    def file_extension(self):
        return os.path.splitext(self.filename)[1][1:]

    @property
    def url(self):
        return self.file.url

    @property
    def sources(self):
        return [
            {
                "src": self.url,
                "type": mimetypes.guess_type(self.filename)[0]
                or "application/octet-stream",
            }
        ]

    def get_usage(self):
        return ReferenceIndex.get_references_to(self).group_by_source_object()

    @property
    def usage_url(self):
        return reverse("wagtailmedia:media_usage", args=(self.id,))

    def is_editable_by_user(self, user):
        from wagtailmedia.permissions import permission_policy

        return permission_policy.user_has_permission_for_instance(user, "change", self)

    def clean(self, *args, **kwargs):
        super().clean(*args, **kwargs)
        if not self.duration:
            self.duration = 0

        if self.thumbnail:
            validate = FileExtensionValidator(ALLOWED_EXTENSIONS_THUMBNAIL)
            validate(self.thumbnail)

        if self.type == "audio" and wagtailmedia_settings.AUDIO_EXTENSIONS:
            validate = FileExtensionValidator(wagtailmedia_settings.AUDIO_EXTENSIONS)
            validate(self.file)
        elif self.type == "video" and wagtailmedia_settings.VIDEO_EXTENSIONS:
            validate = FileExtensionValidator(wagtailmedia_settings.VIDEO_EXTENSIONS)
            validate(self.file)

    class Meta:
        abstract = True
        verbose_name = _("media")


class Media(AbstractMedia):
    pass


def get_media_model():
    from django.apps import apps

    from wagtailmedia.settings import wagtailmedia_settings

    try:
        app_label, model_name = wagtailmedia_settings.MEDIA_MODEL.split(".")
    except AttributeError:
        return Media
    except ValueError:
        raise ImproperlyConfigured(
            "WAGTAILMEDIA[\"MEDIA_MODEL\"] must be of the form 'app_label.model_name'"
        )

    media_model = apps.get_model(app_label, model_name)
    if media_model is None:
        raise ImproperlyConfigured(
            "WAGTAILMEDIA[\"MEDIA_MODEL\"] refers to model '%s' that has not been installed"
            % wagtailmedia_settings.MEDIA_MODEL
        )

    return media_model


# Provides `request` as an argument
media_served = Signal()



from django import VERSION as DJANGO_VERSION
from django.db import migrations


def add_media_permissions_to_admin_groups(apps, schema_editor):
    ContentType = apps.get_model("contenttypes.ContentType")
    Permission = apps.get_model("auth.Permission")
    Group = apps.get_model("auth.Group")

    # Get media permissions
    media_content_type, _created = ContentType.objects.get_or_create(
        model="media",
        app_label="wagtailmedia",
        defaults={"name": "media"} if DJANGO_VERSION < (1, 8) else {},
    )

    add_media_permission, _created = Permission.objects.get_or_create(
        content_type=media_content_type,
        codename="add_media",
        defaults={"name": "Can add media"},
    )
    change_media_permission, _created = Permission.objects.get_or_create(
        content_type=media_content_type,
        codename="change_media",
        defaults={"name": "Can change media"},
    )
    delete_media_permission, _created = Permission.objects.get_or_create(
        content_type=media_content_type,
        codename="delete_media",
        defaults={"name": "Can delete media"},
    )

    # Assign it to Editors and Moderators groups
    for group in Group.objects.filter(name__in=["Editors", "Moderators"]):
        group.permissions.add(
            add_media_permission, change_media_permission, delete_media_permission
        )


def remove_media_permissions(apps, schema_editor):
    """Reverse the above additions of permissions."""
    ContentType = apps.get_model("contenttypes.ContentType")
    Permission = apps.get_model("auth.Permission")
    media_content_type = ContentType.objects.get(
        model="media",
        app_label="wagtailmedia",
    )
    # This cascades to Group
    Permission.objects.filter(
        content_type=media_content_type,
        codename__in=("add_media", "change_media", "delete_media"),
    ).delete()


class Migration(migrations.Migration):
    dependencies = [
        ("wagtailmedia", "0001_initial"),
        # Need to run wagtailcores initial data migration to make sure the groups are created
        ("wagtailcore", "0002_initial_data"),
    ]

    operations = [
        migrations.RunPython(
            add_media_permissions_to_admin_groups, remove_media_permissions
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.paginator import Paginator
from django.forms.utils import flatatt
from django.utils.html import format_html, format_html_join
from django.utils.translation import gettext_lazy as _


if TYPE_CHECKING:
    from .models import AbstractMedia

DEFAULT_PAGE_KEY = "p"


def paginate(request, items, page_key=DEFAULT_PAGE_KEY, per_page=20):
    paginator = Paginator(items, per_page)
    page = paginator.get_page(request.GET.get(page_key))
    return paginator, page


def format_audio_html(item: AbstractMedia) -> str:
    return format_html(
        "<audio controls>\n{sources}\n<p>{fallback}</p>\n</audio>",
        sources=format_html_join(
            "\n", "<source{0}>", [[flatatt(s)] for s in item.sources]
        ),
        fallback=_("Your browser does not support the audio element."),
    )


def format_video_html(item: AbstractMedia) -> str:
    return format_html(
        "<video controls>\n{sources}\n<p>{fallback}</p>\n</video>",
        sources=format_html_join(
            "\n", "<source{0}>", [[flatatt(s)] for s in item.sources]
        ),
        fallback=_("Your browser does not support the video element."),
    )



from django.urls import include, path, reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from wagtail import hooks
from wagtail.admin.menu import MenuItem
from wagtail.admin.search import SearchArea
from wagtail.admin.site_summary import SummaryItem
from wagtail.admin.staticfiles import versioned_static

from wagtailmedia import admin_urls
from wagtailmedia.forms import GroupMediaPermissionFormSet
from wagtailmedia.models import get_media_model
from wagtailmedia.permissions import permission_policy


@hooks.register("register_admin_urls")
def register_admin_urls():
    return [
        path("media/", include((admin_urls, "wagtailmedia"), namespace="wagtailmedia")),
    ]


class MediaMenuItem(MenuItem):
    def is_shown(self, request):
        return permission_policy.user_has_any_permission(
            request.user, ["add", "change", "delete"]
        )


@hooks.register("register_admin_menu_item")
def register_media_menu_item():
    return MediaMenuItem(
        _("Media"),
        reverse("wagtailmedia:index"),
        name="media",
        icon_name="media",
        order=300,
    )


class MediaSummaryItem(SummaryItem):
    order = 300
    template_name = "wagtailmedia/homepage/site_summary_media.html"

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        context["total_media"] = get_media_model().objects.count()
        return context

    def is_shown(self):
        return permission_policy.user_has_any_permission(
            self.request.user, ["add", "change", "delete"]
        )


@hooks.register("construct_homepage_summary_items")
def add_media_summary_item(request, items):
    items.append(MediaSummaryItem(request))


class MediaSearchArea(SearchArea):
    def is_shown(self, request):
        return permission_policy.user_has_any_permission(
            request.user, ["add", "change", "delete"]
        )


@hooks.register("register_admin_search_area")
def register_media_search_area():
    return MediaSearchArea(
        _("Media"),
        reverse("wagtailmedia:index"),
        name="media",
        classnames="icon icon-media",
        order=400,
    )


@hooks.register("register_group_permission_panel")
def register_media_permissions_panel():
    return GroupMediaPermissionFormSet


@hooks.register("describe_collection_contents")
def describe_collection_media(collection):
    media_count = get_media_model().objects.filter(collection=collection).count()
    if media_count:
        url = reverse("wagtailmedia:index") + ("?collection_id=%d" % collection.id)
        return {
            "count": media_count,
            "count_text": ngettext(
                "%(count)s media file", "%(count)s media files", media_count
            )
            % {"count": media_count},
            "url": url,
        }


@hooks.register("insert_editor_css")
def add_media_css_tweaks():
    return format_html(
        '<link rel="stylesheet" href="{}">',
        versioned_static("wagtailmedia/css/wagtailmedia.css"),
    )


@hooks.register("insert_global_admin_css")
def add_media_comparison_css():
    return format_html(
        '<link rel="stylesheet" href="{}">',
        versioned_static("wagtailmedia/css/wagtailmedia-comparison.css"),
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.template.loader import render_to_string

from wagtail.admin.compare import ForeignObjectComparison
from wagtail.admin.panels import FieldPanel

from .models import MediaType
from .utils import format_audio_html, format_video_html
from .widgets import AdminAudioChooser, AdminMediaChooser, AdminVideoChooser


if TYPE_CHECKING:
    from .models import AbstractMedia


class MediaChooserPanel(FieldPanel):
    object_type_name = "media"

    def __init__(self, field_name, media_type=None, *args, **kwargs):
        super().__init__(field_name=field_name, *args, **kwargs)

        self.media_type = media_type

    def clone_kwargs(self):
        kwargs = super().clone_kwargs()
        kwargs.update(media_type=self.media_type)
        return kwargs

    @property
    def _widget_class(self):
        if self.media_type == "audio":
            return AdminAudioChooser
        elif self.media_type == "video":
            return AdminVideoChooser
        return AdminMediaChooser

    def get_form_options(self) -> dict:
        opts = super().get_form_options()
        if "widgets" in opts:
            opts["widgets"][self.field_name] = self._widget_class
        else:
            opts["widgets"] = {self.field_name: self._widget_class}

        return opts


class MediaFieldComparison(ForeignObjectComparison):
    def htmldiff(self) -> str:
        media_item_a, media_item_b = self.get_objects()
        if not all([media_item_a, media_item_b]):
            return ""

        return render_to_string(
            "wagtailmedia/widgets/compare.html",
            {
                "media_item_a": self.render_media_item(media_item_a),
                "media_item_b": self.render_media_item(media_item_b),
            },
        )

    @staticmethod
    def render_media_item(item: AbstractMedia) -> str:
        if item.type == MediaType.AUDIO:
            return format_audio_html(item)
        else:
            return format_video_html(item)



from wagtail.api.v2.serializers import BaseSerializer
from wagtail.api.v2.utils import get_full_url

import rest_framework.fields

from rest_framework.fields import ReadOnlyField


class MediaDownloadUrlField(ReadOnlyField):
    """
    Serializes the "download_url" field for media items.

    Example:
    "download_url": "http://api.example.com/media/my_video.mp4"
    """

    def get_attribute(self, instance):
        return instance

    def to_representation(self, instance):
        return get_full_url(self.context["request"], instance.url)


class MediaItemSerializer(BaseSerializer):
    download_url = MediaDownloadUrlField()
    media_type = rest_framework.fields.CharField(source="type")
